==> tdns-0.0.0.1.tar.gz: tdns-0.0.0.1/tdns.py <==
#!/usr/bin/python

import socket, cdb
import parser
from parserconst import *
from cdbparser import parse_entry
import sys, os, pwd

DBFILENAME = "data.cdb"

ADDR_FAMILY = socket.AF_INET6
AXFR_ALLOWED = []

def build_response(req):
    return [[req[HEADER][ID], 1, req[HEADER][OPCODE], 0, 0, req[HEADER][RD],
             0, 0, RCODE_SUCCESS, req[HEADER][QDCOUNT], 0, 0, 0], 
            req[QUESTION], [], [], []]


def get_addr_lo(db, srcaddr):
    addr = socket.inet_pton(ADDR_FAMILY, srcaddr[0])
    if ADDR_FAMILY == socket.AF_INET6:
        if srcaddr[0].startswith("::ffff:"):
            # Convert to IPv4 address
            addr = socket.inet_pton(socket.AF_INET, srcaddr[0][7:])
        elif srcaddr[0].startswith("2002:"):
            # 6to4 address
            addr = addr[2:6]
        else:
            # TODO: support for IPv6 lo entries
            return db.get("\x00%")
    elif ADDR_FAMILY != socket.AF_INET:
        return db.get("\x00%")
    
    for i in range(len(addr)-1, -1, -1):
        c = db.get("\x00%" + addr[:i])
        if c is not None:
            return c
    return None

def do_axfr_lookup(qname, srcaddr):
    responses = do_lookup(qname, TYPE_SOA, srcaddr, recurse=False)
    if not responses:
        return None
    db = cdb.init(DBFILENAME)
    qname = tuple(qname)
    qnl = len(qname)
    names = [qname]
    for k in db.keys():
        if k[0] == "\0":
            continue
        kp = parser.parse_name(k)
        if kp != qname and kp[-qnl:] == qname:
            names.append(kp)
    responses = []
    for n in names:
        responses += do_lookup(n, QTYPE_ANY, srcaddr, recurse=False)

    # Put the SOA record at the top of the response (dig seems to expect this)
    for i, r in enumerate(responses):
        if r[NAME] == qname and r[TYPE] == TYPE_SOA:
            responses = [r] + responses[:i-1] + responses[i+1:]
            break
    return responses
    

def do_lookup(qname, qtype, srcaddr, recurse=True):
    db = cdb.init(DBFILENAME)
    entries = db.getall(parser.build_name(qname))
    responses = []
    mylo = get_addr_lo(db, srcaddr)
    for entry in entries:
        type, loflag, lo, ttl, timestamp, rdlength, rdata = parse_entry(entry)
        if qtype != QTYPE_ANY and type != qtype and type != TYPE_CNAME:
            continue
        if lo is not None and lo != mylo:
            continue
        # TODO - implement timestamp settings
        responses.append((qname, type, CLASS_IN, ttl, rdlength, rdata))
    if recurse and not responses:
        # DNS and case insensitivity
        # See http://www.maradns.org/case_insensitivity.html
        qnlowercase = tuple([x.lower() for x in qname])
        if qnlowercase != qname:
            return do_lookup(qnlowercase, qtype, srcaddr, recurse=False)
    return responses

def handle_request(req, srcaddr):
    # Preliminary filtering (deal with DNS's quirkiness).
    # Questions:
    # - What should we do when qdcount > 1
    # - What should we do when qdcount == 0
    # - Should we pay attention to qclass at all?
    if req[HEADER][QR]:
        # Not a query; drop the packet
        return
    if req[HEADER][OPCODE] != 0 or req[HEADER][QDCOUNT] > 1:
        # Not implemented error
        # FIXME - should the AA bit be set here?
        rsp = build_response(req)
        rsp[HEADER][RCODE] = RCODE_ENOTIMPL
        return rsp
    
    if req[HEADER][QDCOUNT] == 0:
        # No question, no answer.
        return
    
    rsp = build_response(req)
    authority = []
    additional = []
    rsp[HEADER][AA] = 1
    q = req[QUESTION][0]
    if q[QCLASS] == QCLASS_ANY:
        # Don't be authoritative for QCLASS=*.  RFC 1035 s.6.2
        rsp[HEADER][AA] = 0
    elif q[QCLASS] != CLASS_IN:
        # Don't respond for non-IN query class
        rsp = build_response(req)
        rsp[HEADER][RCODE] = RCODE_ENOTIMPL
        return rsp
    
    if q[QTYPE] == QTYPE_AXFR:
        if q[QNAME] in AXFR_ALLOWED:
            answer = do_axfr_lookup(q[QNAME], srcaddr)
        else:
            qnlowercase = tuple([x.lower() for x in q[QNAME]])
            if qnlowercase != q[QNAME]:
                if qnlowercase in AXFR_ALLOWED:
                    answer = do_axfr_lookup(qnlowercase, srcaddr)
                else:
                    rsp[HEADER][RCODE] = RCODE_EREFUSED
                    return rsp
            else:
                rsp[HEADER][RCODE] = RCODE_EREFUSED
                return rsp
    else:
        answer = do_lookup(q[QNAME], q[QTYPE], srcaddr)
    if not answer:
        rsp[HEADER][RCODE] = RCODE_ENXDOMAIN
        return rsp

    # Do CNAME lookups
    for x in answer:
        if x[TYPE] == TYPE_CNAME:
            try:
                qname = parser.parse_name(x[RDATA])
            except (ValueError, TypeError):
                pass
            additional += do_lookup(qname, q[QTYPE], srcaddr, recurse=False)
    
    rsp[HEADER][ANCOUNT], rsp[ANSWER] = len(answer), tuple(answer)
    rsp[HEADER][NSCOUNT], rsp[AUTHORITY] = len(authority), tuple(authority)
    rsp[HEADER][ARCOUNT], rsp[ADDITIONAL] = len(additional), tuple(additional)
    return rsp
    
def udpserver():
    skt = socket.socket(ADDR_FAMILY, socket.SOCK_DGRAM)
    skt.bind(('::ffff:10.159.17.101', 53))
    #skt.bind(('', 5353))
    
    if os.geteuid() == 0 or os.getuid() == 0:
        # Drop root privs
        uid, gid = pwd.getpwnam("nobody")[2:4]
        os.setgid(gid)
        os.setuid(uid)
    
    while True:
        pkt, addr = skt.recvfrom(512)
        try:
            req = parser.parse_message(pkt)
            #print "*** Request:"
            #parser.pretty_print(req)
            response = handle_request(req, addr)
            #print "*** Response:"
            if response is not None:
                #parser.pretty_print(response)
                msg = parser.build_message(response)
                if len(msg) > 512:
                    response[HEADER][TC] = 1
                    msg = parser.build_message(response)
                skt.sendto(msg, addr)
            else:
                pass
                #print "None\n"
        except:
            raise
#            try:
#                print >>sys.stderr, "Error constructing response."
#                response = build_response(req)
#                response[HEADER][RCODE] = RCODE_ESERVER
#                skt.sendto(parser.build_message(response), addr)
#            except:
#                print >>sys.stderr, " Error sending error message!"
#                pass

def tcphandler():
    # TODO: handle AXFR
    global ADDR_FAMILY, AXFR_ALLOWED
    if os.geteuid() == 0 or os.getuid() == 0:
        # Drop root privs
        uid, gid = pwd.getpwnam("nobody")[2:4]
        os.setgid(gid)
        os.setuid(uid)
    
    AXFR_ALLOWED = [tuple(x.split(".")) for x in os.environ.get("AXFR", "").split("/")]
    
    if os.environ.get("PROTO") == "TCP":
        ADDR_FAMILY = socket.AF_INET
    elif os.environ.get("PROTO") == "TCP6":
        ADDR_FAMILY = socket.AF_INET6
    else:
        print >>sys.stderr, "Unrecognized $PROTO value: " + `os.environ.get("PROTO")`
        sys.exit(1)
    addr = (os.environ["TCPREMOTEIP"], None)
    
    while True:
        try:
            pktlen = ord(sys.stdin.read(1)) << 8 | ord(sys.stdin.read(1))
        except TypeError:
            # Connection closed
            break
        pkt = sys.stdin.read(pktlen)
        try:
            req = parser.parse_message(pkt)
            #print >>sys.stderr, "*** Request:"
            #print >>sys.stderr, parser.pretty_format(req)
            response = handle_request(req, addr)
            #print >>sys.stderr, "*** Response:"
            if response is not None:
                #print >>sys.stderr, parser.pretty_format(response)
                msg = parser.build_message(response)
                if len(msg) > 0xffff:
                    print >>sys.stderr, "Internal error: message too long"
                    sys.exit(1)
                sys.stdout.write(chr(len(msg) >> 8) + chr(len(msg) & 0xff))
                sys.stdout.write(msg)
                sys.stdout.flush()
            else:
                response = parser.build_response(req)
                response[HEADER][RCODE] = RCODE_ENXDOMAIN
                response[HEADER][AA] = 0
                if len(msg) > 0xffff:
                    print >>sys.stderr, "Internal error: message too long"
                    sys.exit(1)
                sys.stdout.write(chr(len(msg) >> 8) + chr(len(msg) & 0xff))
                sys.stdout.write(msg)
                sys.stdout.flush()
        except:
            raise
#            try:
#                print >>sys.stderr, "Error constructing response."
#                response = build_response(req)
#                response[HEADER][RCODE] = RCODE_ESERVER
#                skt.sendto(parser.build_message(response), addr)
#            except:
#                print >>sys.stderr, " Error sending error message!"
#                pass

if __name__ == '__main__':
    if not sys.argv[1:]:
        print >>sys.stderr, "Usage: %s udpserver|tcphandler" % sys.argv[0]
        sys.exit(1)
    if sys.argv[1] == "udpserver":
        udpserver()
    elif sys.argv[1] == "tcphandler":
        tcphandler()
    else:
        print >>sys.stderr, "Usage: %s udpserver|tcphandler" % sys.argv[0]
        sys.exit(1)

==> tdns-0.0.0.2.tar.gz: tdns-0.0.0.2/tdns.py <==
#!/usr/bin/python

import socket, cdb
import parser
from parserconst import *
from cdbparser import parse_entry
import sys, os, pwd

DBFILENAME = "data.cdb"

ADDR_FAMILY = socket.AF_INET6
AXFR_ALLOWED = []

def build_response(req):
    return [[req[HEADER][ID], 1, req[HEADER][OPCODE], 0, 0, req[HEADER][RD],
             0, 0, RCODE_SUCCESS, req[HEADER][QDCOUNT], 0, 0, 0], 
            req[QUESTION], [], [], []]


def get_addr_lo(db, srcaddr):
    addr = socket.inet_pton(ADDR_FAMILY, srcaddr[0])
    if ADDR_FAMILY == socket.AF_INET6:
        if srcaddr[0].startswith("::ffff:"):
            # Convert to IPv4 address
            addr = socket.inet_pton(socket.AF_INET, srcaddr[0][7:])
        elif srcaddr[0].startswith("2002:"):
            # 6to4 address
            addr = addr[2:6]
        else:
            # TODO: support for IPv6 lo entries
            return db.get("\x00%")
    elif ADDR_FAMILY != socket.AF_INET:
        return db.get("\x00%")
    
    for i in range(len(addr)-1, -1, -1):
        c = db.get("\x00%" + addr[:i])
        if c is not None:
            return c
    return None

def do_axfr_lookup(qname, srcaddr):
    responses = do_lookup(qname, TYPE_SOA, srcaddr, recurse=False)
    if not responses:
        return None
    db = cdb.init(DBFILENAME)
    qname = tuple(qname)
    qnl = len(qname)
    names = [qname]
    for k in db.keys():
        if k[0] == "\0":
            continue
        kp = parser.parse_name(k)
        if kp != qname and kp[-qnl:] == qname:
            names.append(kp)
    responses = []
    for n in names:
        responses += do_lookup(n, QTYPE_ANY, srcaddr, recurse=False)

    # Put the SOA record at the beginning *and* at the end of the response
    # (dig seems to expect this)
    for i, r in enumerate(responses):
        if r[NAME] == qname and r[TYPE] == TYPE_SOA:
            responses = [r] + responses[:i-1] + responses[i+1:] + [r]
            break
    return responses
    

def do_lookup(qname, qtype, srcaddr, recurse=True):
    db = cdb.init(DBFILENAME)
    entries = db.getall(parser.build_name(qname))
    responses = []
    mylo = get_addr_lo(db, srcaddr)
    for entry in entries:
        type, loflag, lo, ttl, timestamp, rdlength, rdata = parse_entry(entry)
        if qtype != QTYPE_ANY and type != qtype and type != TYPE_CNAME:
            continue
        if lo is not None and lo != mylo:
            continue
        # TODO - implement timestamp settings
        responses.append((qname, type, CLASS_IN, ttl, rdlength, rdata))
    if recurse and not responses:
        # DNS and case insensitivity
        # See http://www.maradns.org/case_insensitivity.html
        qnlowercase = tuple([x.lower() for x in qname])
        if qnlowercase != qname:
            return do_lookup(qnlowercase, qtype, srcaddr, recurse=False)
    return responses

def handle_request(req, srcaddr):
    # Preliminary filtering (deal with DNS's quirkiness).
    # Questions:
    # - What should we do when qdcount > 1
    # - What should we do when qdcount == 0
    # - Should we pay attention to qclass at all?
    if req[HEADER][QR]:
        # Not a query; drop the packet
        return
    if req[HEADER][OPCODE] != 0 or req[HEADER][QDCOUNT] > 1:
        # Not implemented error
        # FIXME - should the AA bit be set here?
        rsp = build_response(req)
        rsp[HEADER][RCODE] = RCODE_ENOTIMPL
        return rsp
    
    if req[HEADER][QDCOUNT] == 0:
        # No question, no answer.
        return
    
    rsp = build_response(req)
    authority = []
    additional = []
    rsp[HEADER][AA] = 1
    q = req[QUESTION][0]
    if q[QCLASS] == QCLASS_ANY:
        # Don't be authoritative for QCLASS=*.  RFC 1035 s.6.2
        rsp[HEADER][AA] = 0
    elif q[QCLASS] != CLASS_IN:
        # Don't respond for non-IN query class
        rsp = build_response(req)
        rsp[HEADER][RCODE] = RCODE_ENOTIMPL
        return rsp
    
    if q[QTYPE] == QTYPE_AXFR:
        if q[QNAME] in AXFR_ALLOWED:
            answer = do_axfr_lookup(q[QNAME], srcaddr)
        else:
            qnlowercase = tuple([x.lower() for x in q[QNAME]])
            if qnlowercase != q[QNAME]:
                if qnlowercase in AXFR_ALLOWED:
                    answer = do_axfr_lookup(qnlowercase, srcaddr)
                else:
                    rsp[HEADER][RCODE] = RCODE_EREFUSED
                    return rsp
            else:
                rsp[HEADER][RCODE] = RCODE_EREFUSED
                return rsp
    else:
        answer = do_lookup(q[QNAME], q[QTYPE], srcaddr)
    if not answer:
        rsp[HEADER][RCODE] = RCODE_ENXDOMAIN
        return rsp

    # Do CNAME lookups
    for x in answer:
        if x[TYPE] == TYPE_CNAME:
            try:
                qname = parser.parse_name(x[RDATA])
            except (ValueError, TypeError):
                pass
            additional += do_lookup(qname, q[QTYPE], srcaddr, recurse=False)
    
    rsp[HEADER][ANCOUNT], rsp[ANSWER] = len(answer), tuple(answer)
    rsp[HEADER][NSCOUNT], rsp[AUTHORITY] = len(authority), tuple(authority)
    rsp[HEADER][ARCOUNT], rsp[ADDITIONAL] = len(additional), tuple(additional)
    return rsp
    
def udpserver():
    skt = socket.socket(ADDR_FAMILY, socket.SOCK_DGRAM)
    skt.bind(('::ffff:10.159.17.101', 53))
    #skt.bind(('', 5353))
    
    if os.geteuid() == 0 or os.getuid() == 0:
        # Drop root privs
        uid, gid = pwd.getpwnam("nobody")[2:4]
        os.setgid(gid)
        os.setuid(uid)
    
    while True:
        pkt, addr = skt.recvfrom(512)
        try:
            req = parser.parse_message(pkt)
            #print "*** Request:"
            #parser.pretty_print(req)
            response = handle_request(req, addr)
            #print "*** Response:"
            if response is not None:
                #parser.pretty_print(response)
                msg = parser.build_message(response)
                if len(msg) > 512:
                    response[HEADER][TC] = 1
                    msg = parser.build_message(response)
                skt.sendto(msg, addr)
            else:
                pass
                #print "None\n"
        except:
            raise
#            try:
#                print >>sys.stderr, "Error constructing response."
#                response = build_response(req)
#                response[HEADER][RCODE] = RCODE_ESERVER
#                skt.sendto(parser.build_message(response), addr)
#            except:
#                print >>sys.stderr, " Error sending error message!"
#                pass

def tcphandler():
    # TODO: handle AXFR
    global ADDR_FAMILY, AXFR_ALLOWED
    if os.geteuid() == 0 or os.getuid() == 0:
        # Drop root privs
        uid, gid = pwd.getpwnam("nobody")[2:4]
        os.setgid(gid)
        os.setuid(uid)
    
    AXFR_ALLOWED = [tuple(x.split(".")) for x in os.environ.get("AXFR", "").split("/")]
    
    if os.environ.get("PROTO") == "TCP":
        ADDR_FAMILY = socket.AF_INET
    elif os.environ.get("PROTO") == "TCP6":
        ADDR_FAMILY = socket.AF_INET6
    else:
        print >>sys.stderr, "Unrecognized $PROTO value: " + `os.environ.get("PROTO")`
        sys.exit(1)
    addr = (os.environ["TCPREMOTEIP"], None)
    
    while True:
        try:
            pktlen = ord(sys.stdin.read(1)) << 8 | ord(sys.stdin.read(1))
        except TypeError:
            # Connection closed
            break
        pkt = sys.stdin.read(pktlen)
        try:
            req = parser.parse_message(pkt)
            #print >>sys.stderr, "*** Request:"
            #print >>sys.stderr, parser.pretty_format(req)
            response = handle_request(req, addr)
            #print >>sys.stderr, "*** Response:"
            if response is not None:
                #print >>sys.stderr, parser.pretty_format(response)
                msg = parser.build_message(response)
                if len(msg) > 0xffff:
                    print >>sys.stderr, "Internal error: message too long"
                    sys.exit(1)
                sys.stdout.write(chr(len(msg) >> 8) + chr(len(msg) & 0xff))
                sys.stdout.write(msg)
                sys.stdout.flush()
            else:
                response = parser.build_response(req)
                response[HEADER][RCODE] = RCODE_ENXDOMAIN
                response[HEADER][AA] = 0
                if len(msg) > 0xffff:
                    print >>sys.stderr, "Internal error: message too long"
                    sys.exit(1)
                sys.stdout.write(chr(len(msg) >> 8) + chr(len(msg) & 0xff))
                sys.stdout.write(msg)
                sys.stdout.flush()
        except:
            raise
#            try:
#                print >>sys.stderr, "Error constructing response."
#                response = build_response(req)
#                response[HEADER][RCODE] = RCODE_ESERVER
#                skt.sendto(parser.build_message(response), addr)
#            except:
#                print >>sys.stderr, " Error sending error message!"
#                pass

if __name__ == '__main__':
    if not sys.argv[1:]:
        print >>sys.stderr, "Usage: %s udpserver|tcphandler" % sys.argv[0]
        sys.exit(1)
    if sys.argv[1] == "udpserver":
        udpserver()
    elif sys.argv[1] == "tcphandler":
        tcphandler()
    else:
        print >>sys.stderr, "Usage: %s udpserver|tcphandler" % sys.argv[0]
        sys.exit(1)
